@@ -21,6 +21,14 @@
 
 import spidev
 import time
+import threading
+import collections
+
+class dummy_context_mgr():
+	def __enter__(self):
+		return None
+	def __exit__(self, exc_type, exc_value, traceback):
+		return False
 
 class Radio:
 	_register_map = [
@@ -222,7 +230,13 @@ class Radio:
 		spi.open(spi_bus, spi_dev)
 		self._spi = spi
 
-		self.configure(config)
+		self._dequeue_thread = None
+		self._last_syncword = None
+
+		self._software_tx_queue = {}
+		self._software_tx_queue_next_time = {}
+
+		self.configure(config, update = False)
 
 		if len(self._register_map) != 53:
 			raise ValueError('Inconsistent register map!')
@@ -230,19 +244,47 @@ class Radio:
 		return None
 
 	def __del__(self):
+		self._debug('Deleting object')
+		self._config['use_software_tx_queue'] = False
 		self._spi.close()
 
 	def _debug(self, message):
-		if self._config is not None and 'debug_log_command' in self._config:
+		if 'debug_log_command' in self._config:
 			self._config['debug_log_command'](message)
 		return None
 
-	def _reset_device(self):
-		if self._config is not None:
-			if self._config['reset_command'] is not None:
-				return self._config['reset_command']()
+	def _error(self, message):
+		log_command = None
+		if 'error_log_command' in self._config:
+			log_command = self._config['error_log_command']
+		elif 'debug_log_command' in self._config:
+			log_command = self._config['debug_log_command']
+		log_command(message)
 		return None
 
+	def _get_mutex(self, real_mutex = True):
+		if not real_mutex:
+			return dummy_context_mgr()
+
+		mutex = self._config.get('mutex', dummy_context_mgr())
+		return mutex
+
+	def _reset_device(self):
+		reset_command = self._config.get('reset_command', None)
+
+		if reset_command is None:
+			return None
+
+		reset_command()
+
+		return None
+
+	def _should_use_queue(self):
+		if 'use_software_tx_queue' in self._config:
+			return self._config['use_software_tx_queue']
+
+		return False
+
 	def _register_name(self, reg_number):
 		return self._register_map[reg_number]['name']
 
@@ -310,6 +352,7 @@ class Radio:
 
 	def _put_register_high_low(self, reg, high, low, delay = 7):
 		reg = self._register_number(reg)
+
 		result = self._spi.xfer([reg, high, low], self._spi.max_speed_hz, delay)
 
 		if reg & 0x80 == 0x80:
@@ -385,19 +428,36 @@ class Radio:
 		# Return the filled in structure
 		return result
 
-	def configure(self, config):
-		self._config = config
-
+	def configure(self, config, update = True):
 		if config is None:
-			return None
+			config = {}
 
-		self._spi.max_speed_hz = self._config.get('frequency', 4000000)
-		self._spi.bits_per_word = self._config.get('bits_per_word', 8)
-		self._spi.cshigh = self._config.get('csigh', False)
-		self._spi.no_cs  = self._config.get('no_cs', False)
-		self._spi.lsbfirst = self._config.get('lsbfirst', False)
-		self._spi.threewire = self._config.get('threewire', False)
-		self._spi.mode = self._config.get('mode', 1)
+		if update:
+			self._config.update(config)
+		else:
+			self._config = config
+
+		with self._get_mutex():
+			self._spi.max_speed_hz = self._config.get('frequency', 4000000)
+			self._spi.bits_per_word = self._config.get('bits_per_word', 8)
+			self._spi.cshigh = self._config.get('csigh', False)
+			self._spi.no_cs  = self._config.get('no_cs', False)
+			self._spi.lsbfirst = self._config.get('lsbfirst', False)
+			self._spi.threewire = self._config.get('threewire', False)
+			self._spi.mode = self._config.get('mode', 1)
+
+		# If using a queue, start a thread to run the queue
+		if self._should_use_queue():
+			if self._dequeue_thread is None:
+				self._dequeue_thread = threading.Thread(target = self._run_queue, daemon = True)
+				self._software_tx_queue_mutex = threading.Lock()
+				self._dequeue_thread.start()
+		else:
+			if self._dequeue_thread is not None:
+				self._debug("Joining existing thread to wait for termination")
+				self._dequeue_thread.join()
+				self._dequeue_thread = None
+				self._software_tx_queue_mutex = None
 
 		return None
 
@@ -418,7 +478,20 @@ class Radio:
 
 		return state
 
-	def set_syncword(self, syncword):
+	def set_syncword(self, syncword, force = False, submit_queue = '__DEFAULT__'):
+		# If queuing is being used, just store this message
+		if submit_queue is not None and self._should_use_queue():
+			self._enqueue(submit_queue, syncword, None, None, post_delay = 0)
+			return None
+
+		# Do not set the syncword again if it's not needed
+		if not force:
+			if self._last_syncword is not None:
+				if syncword == self._last_syncword:
+					return None
+
+		self._last_syncword = syncword
+
 		packet_config = self.get_register_bits('packet_config')
 		packet_config['syncword_len'] = len(syncword) - 1
 
@@ -443,7 +516,7 @@ class Radio:
 
 		return None
 
-	def fill_fifo(self, message, include_length = True):
+	def fill_fifo(self, message, include_length = True, lock = True):
 		new_message = [self._register_number('fifo')]
 		if include_length:
 			new_message = new_message + [len(message)]
@@ -451,59 +524,214 @@ class Radio:
 		log_message = [] + new_message
 
 		# Transfer the message
-		result = self._spi.xfer(new_message, self._spi.max_speed_hz, 10)
+		with self._get_mutex(lock):
+			result = self._spi.xfer(new_message, self._spi.max_speed_hz, 0)
+
 		self._debug("Writing: {} = {}".format(log_message, result))
 
 		return new_message
 
-	def transmit(self, message, channel = None):
-		if channel is None:
-			state = self.get_register_bits('radio_state')
-			channel = state['channel']
+	def transmit(self, message, channel = None, lock = True, post_delay = 0, syncword = None, submit_queue = '__DEFAULT__'):
+		# If we are using a radio transmit queue, just queue this message
+		# (unless we are called from the dequeue procedure)
+		if submit_queue is not None and self._should_use_queue():
+			if syncword is None:
+				syncword = self._last_syncword
+			self._enqueue(submit_queue, syncword, message, channel, post_delay = post_delay)
+			return True
 
-		# Initialize the transmitter
-		self.put_register_bits('radio_state', {
-			'tx_enabled': 0,
-			'rx_enabled': 0,
-			'channel': 0
-		})
+		sent_packet = True
 
-		self.put_register_bits('fifo_state', {
-			'clear_read': 1,
-			'clear_write': 1
-		})
+		with self._get_mutex(lock):
+			# Set the syncword
+			if syncword is not None:
+				self.set_syncword(syncword, submit_queue = None)
 
-		# Format message to send to fifo
-		self.fill_fifo(message, True)
+			if channel is None:
+				state = self.get_register_bits('radio_state')
+				channel = state['channel']
 
-		# Tell the radio to transmit the FIFO buffer to the specified channel
-		self.put_register_bits('radio_state', {
-			'tx_enabled': 1,
-			'rx_enabled': 0,
-			'channel': channel
-		})
+			# Initialize the transmitter
+			self.put_register_bits('radio_state', {
+				'tx_enabled': 0,
+				'rx_enabled': 0,
+				'channel': 0
+			})
 
-		# Wait for buffer to empty
-		# XXX: Untested
-		while True:
-			radio_status = self.get_register_bits('status')
-			self._debug("radio_status={}".format(radio_status))
+			self.put_register_bits('fifo_state', {
+				'clear_read': 1,
+				'clear_write': 1
+			})
 
-			if radio_status['packet_flag'] == 1:
-				break
-			time.sleep(0.001)
+			# Format message to send to fifo
+			self.fill_fifo(message, include_length = True, lock = False)
 
-		return True
+			# Tell the radio to transmit the FIFO buffer to the specified channel
+			self.put_register_bits('radio_state', {
+				'tx_enabled': 1,
+				'rx_enabled': 0,
+				'channel': channel
+			})
 
-	def multi_transmit(self, message, channels, retries = 3, delay = 0.1):
-		for channel in channels:
+			# Wait for buffer to empty
+			# XXX: Untested
+			while True:
+				radio_status = self.get_register_bits('status')
+				self._debug("radio_status={}".format(radio_status))
+
+				if radio_status['packet_flag'] == 1:
+					break
+
+				if radio_status['framer_status'] == 0:
+					sent_packet = False
+					break
+				time.sleep(0.001)
+
+		if post_delay != 0:
+			time.sleep(post_delay)
+
+		return sent_packet
+
+	def multi_transmit(self, message, channels, retries = 3, delay = 0.1, syncword = None, submit_queue = '__DEFAULT__'):
+		if len(channels) == 0 or retries == 0:
+			self._error("Asked to send the message {} a total of zero times ({} channels, {} retries)".format(message, channels, retries))
+
+		for channel_idx in range(len(channels)):
+			if channel_idx == (len(channels) - 1):
+				retries -= 1
+			channel = channels[channel_idx]
 			for i in range(retries):
-				if not self.transmit(message, channel):
+				if not self.transmit(message, channel, post_delay = 0, syncword = syncword, submit_queue = submit_queue):
 					return False
-				if delay != 0:
-					time.sleep(delay / retries)
+		if not self.transmit(message, channel, post_delay = delay, syncword = syncword, submit_queue = submit_queue):
+			return False
+
 		return True
 
+	def _enqueue(self, submit_queue, syncword, message, channel, post_delay = 0):
+		if not self._should_use_queue():
+			raise ValueError('internal error: _enqueue called with queueing disabled')
+
+		with self._software_tx_queue_mutex:
+			if submit_queue not in self._software_tx_queue:
+				self._software_tx_queue[submit_queue] = collections.deque([])
+
+			self._software_tx_queue[submit_queue].append({
+				'syncword': syncword,
+				'message': message,
+				'channel': channel,
+				'post_delay': post_delay
+			})
+
+		return None
+
+	def _run_queue(self):
+		self._debug("Started run_queue process")
+
+		sleep_time = 0
+		while True:
+			if sleep_time != 0:
+				self._debug("Sleeping for {} seconds".format(sleep_time))
+				time.sleep(sleep_time)
+
+			with self._software_tx_queue_mutex:
+				for queue in self._software_tx_queue:
+					if len(self._software_tx_queue[queue]) != 0:
+						self._debug("Running the queue named {}: {} items left".format(queue, len(self._software_tx_queue[queue])))
+
+			try:
+				[processed_items, remaining_items] = self._run_queue_once()
+			except Exception as error_info:
+				self._error("Failed to run queue: {}".format(error_info.args))
+				processed_items = 0
+				remaining_items = 0
+
+			self._debug("Completed running the queue, did {} items and {} items left (continue queue = {})".format(processed_items, remaining_items, self._should_use_queue()))
+
+			if remaining_items == 0:
+				# If the queue is empty and we are no longer queuing
+				# events, exit this function (which should be joined)
+				if not self._should_use_queue():
+					self._debug("Request to stop run_queue process, exiting")
+					return None
+
+				# If there are no events, wait a bit
+				# longer before trying again
+				sleep_time = 0.5
+				continue
+
+			if processed_items == 0:
+				# If we processed no items, but there are items left to
+				# process, back off slightly
+				if sleep_time == 0:
+					sleep_time = 0.001
+				else:
+					sleep_time = min(0.5, sleep_time * 2)
+				continue
+
+			# If there are more events to process, try again
+			sleep_time = 0
+
+		return None
+
+	def _run_queue_once(self):
+		to_transmit = []
+		remaining_items = 0
+		now = time.time()
+		with self._software_tx_queue_mutex:
+			for submit_queue in self._software_tx_queue:
+				# Determine if we should run this queue yet
+				if submit_queue not in self._software_tx_queue_next_time:
+					self._software_tx_queue_next_time[submit_queue] = now
+
+				queue_next_time = self._software_tx_queue_next_time[submit_queue]
+				if now < queue_next_time:
+					remaining_items += len(self._software_tx_queue[submit_queue])
+					continue
+
+				# Record how many items to pop off this queue
+				pop_items = 0
+				for item in self._software_tx_queue[submit_queue]:
+					pop_items += 1
+
+					# If the last item we're about to transmit requires a delay, make
+					# a note of it in the queue time and don't pull anything else
+					# from this queue
+					self._software_tx_queue_next_time[submit_queue] = now + item['post_delay']
+					if item['post_delay'] != 0:
+						break
+
+				# Pop off the items to transmit in this run into a list
+				if pop_items != 0:
+					self._debug("Found {} items to transmit in the {} queue".format(pop_items, submit_queue))
+				while pop_items != 0:
+					to_transmit.append(self._software_tx_queue[submit_queue].popleft())
+					pop_items -= 1
+
+				remaining_items += len(self._software_tx_queue[submit_queue])
+
+		default_syncword = None
+
+		self._debug("Getting ready to transmit {} items".format(len(to_transmit)))
+		with self._get_mutex():
+			for item in to_transmit:
+				self._debug("Transmitting item {}".format(item))
+				syncword = item['syncword']
+				message = item['message']
+				channel = item['channel']
+
+				if syncword is not None:
+					default_syncword = syncword
+				else:
+					syncword = default_syncword
+
+				if message is None or channel is None:
+					continue
+
+				self.transmit(message, channel, lock = False, submit_queue = None, syncword = syncword, post_delay = 0)
+
+		return [len(to_transmit), remaining_items]
+		
 	def start_listening(self, channel):
 		# Initialize the receiver
 		self.stop_listening()
@@ -533,56 +761,58 @@ class Radio:
 		return True
 
 	def receive(self, channel = None, wait = False, length = None, wait_time = 0.1):
-		if wait:
-			if channel is None:
-				state = self.get_register_bits('radio_state')
-				channel = state['channel']
+		with self._get_mutex():
+			if wait:
+				if channel is None:
+					state = self.get_register_bits('radio_state')
+					channel = state['channel']
 
-			self.start_listening(channel)
+				self.start_listening(channel)
 
-		message = []
+			message = []
 
-		while True:
-			radio_status = self.get_register_bits('status')
-			self._debug("radio_status={}".format(radio_status))
-			if radio_status['packet_flag'] == 0:
-				if wait:
-					time.sleep(wait_time)
+			while True:
+				radio_status = self.get_register_bits('status')
+				self._debug("radio_status={}".format(radio_status))
+				if radio_status['packet_flag'] == 0:
+					if wait:
+						time.sleep(wait_time)
+						continue
+					else:
+						self._unlock_radio()
+						return None
+
+				if radio_status['crc_error'] == 1:
+					# Handle invalid packet ?
+					self.start_listening(channel)
 					continue
-				else:
-					return None
 
-			if radio_status['crc_error'] == 1:
-				# Handle invalid packet ?
-				self.start_listening(channel)
-				continue
-
-			# Data is available, read it from the FIFO register
-			# The first result will include the length
-			# XXX *IF* length encoding is enabled ?
-			fifo_data = self.get_register('fifo')
-			message_length = fifo_data >> 8
-
-			if message_length == 0:
-				self.start_listening(channel)
-				continue
-
-			# Keep track of the total message length to truncate it
-			final_message_length = message_length
-
-			message += [fifo_data & 0xff]
-			message_length -= 1
-
-			# Read subsequent bytes from the FIFO register until
-			# there are no more bytes to read
-			while message_length > 0:
+				# Data is available, read it from the FIFO register
+				# The first result will include the length
+				# XXX *IF* length encoding is enabled ?
 				fifo_data = self.get_register('fifo')
-				message += [fifo_data >> 8, fifo_data & 0xff]
-				message_length -= 2
+				message_length = fifo_data >> 8
 
-			# Truncate the message to its final size, since we have
-			# to read in 16-bit words, we may have an extra byte
-			message = message[0:final_message_length]
-			break
+				if message_length == 0:
+					self.start_listening(channel)
+					continue
+
+				# Keep track of the total message length to truncate it
+				final_message_length = message_length
+
+				message += [fifo_data & 0xff]
+				message_length -= 1
+
+				# Read subsequent bytes from the FIFO register until
+				# there are no more bytes to read
+				while message_length > 0:
+					fifo_data = self.get_register('fifo')
+					message += [fifo_data >> 8, fifo_data & 0xff]
+					message_length -= 2
+
+				# Truncate the message to its final size, since we have
+				# to read in 16-bit words, we may have an extra byte
+				message = message[0:final_message_length]
+				break
 
 		return message
